import string
import re
import random
import datetime
import base64
import os
import json
import copy
from collections import abc


class AString(str):
    """AString represents "Aries String", a sub-class of python built-in str with additional methods.
    AString inherits all methods of the python str.
    Instance of AString can be use in place of python str.

    Important: AString converts NoneType to empty string.

    For methods in python str returning a str, list, or tuple,
        additional post-processing are added to convert the returning str values to instances AString.
        e.g., AString("hello").title() will return AString("Hello").
    This is designed to enable method chaining for AString methods,
        e.g., AString("hello").title().append_today().

    """
    def __new__(cls, string_literal):
        """Creates a new string from string literal.
        str is immutable and it cannot be modified in __init__

        Args:
            string_literal (str): String literal

        Returns:
            str: [description]
        """
        if string_literal is None:
            return super(AString, cls).__new__(cls, "")
        else:
            return super(AString, cls).__new__(cls, string_literal)

    def __getattribute__(self, item):
        """Wraps the existing methods of python str to return AString objects instead of build-in strings.
        If the existing method returns a str, it will be converted to an instance of AString.
        If the existing method returns a list or tuple,
            the str values in the list or tuple will be converted to instances of AString.

        See Also:
            https://docs.python.org/3.5/reference/datamodel.html#object.__getattribute__
            https://stackoverflow.com/questions/7255655/how-to-subclass-str-in-python

        """
        def method(s, *args, **kwargs):
            # super() returns a a proxy object that delegates method calls to a parent or sibling class
            # See https://docs.python.org/3/library/functions.html#super
            value = getattr(super(AString, self), item)(*args, **kwargs)
            # Return value is str, list, tuple:
            if isinstance(value, str):
                return type(s)(value)
            elif isinstance(value, list):
                return [type(s)(i) for i in value]
            elif isinstance(value, tuple):
                return tuple(type(s)(i) for i in value)
            # Return value is dict, bool, or int
            return value

        # If the method is a method of str
        if item in dir(str):  # only handle str methods here
            # Bound method
            return method.__get__(self, type(self))
        else:
            # Delegate to parent
            return super(AString, self).__getattribute__(item)

    def prepend(self, s, delimiter='_'):
        """Prepends the string with another string or a list of strings, connected by the delimiter.

        Args:
            s (str/list): A string or a list of strings to be prepended to the filename.
            delimiter: A string concatenating the original filename and each of the prepended strings.

        Returns: An AString instance

        """
        if not isinstance(s, list):
            s = [s]
        return AString(delimiter.join(s + [self]))

    def append(self, s, delimiter='_'):
        """Appends a list of strings, connected by the delimiter.

        Args:
            s (str/list): A string or a list of strings to be appended to the filename.
            delimiter: A string concatenating the original filename and each of the appended strings.

        Returns: An AString instance

        """
        if not isinstance(s, list):
            s = [s]
        return AString(delimiter.join([self] + s))

    def append_datetime(self, dt=datetime.datetime.now(), fmt="%Y%m%d_%H%M%S"):
        """Appends date and time.
        The current date and time will be appended by default.

        Args:
            dt (datetime.datetime): A datetime.datetime instance.
            fmt (str): The format of the datetime.

        Returns: An AString instance

        """
        datetime_string = dt.strftime(fmt)
        return self.append(datetime_string)

    def append_today(self, fmt="%Y%m%d"):
        """Appends today's date.

        Args:
            fmt (str): The format of the date.

        Returns: An AString instance

        """
        return self.append_datetime(fmt=fmt)

    def append_random(self, choices, n):
        """Appends a random string of n characters.

        Args:
            choices (str): A string including the choices of characters.
            n (int): The number of characters to be appended.

        Returns: An AString instance

        """
        random_chars = ''.join(random.choice(choices) for _ in range(n))
        return self.append(random_chars)

    def append_random_letters(self, n):
        """Appends a random string of letters.

        Args:
            n (int): The number of characters to be appended.

        Returns: An AString instance

        """
        return self.append_random(string.ascii_letters, n)

    def append_random_uppercase(self, n):
        """Appends a random string of uppercase letters.

        Args:
            n (int): The number of characters to be appended.

        Returns: An AString instance

        """
        return self.append_random(string.ascii_uppercase, n)

    def append_random_lowercase(self, n):
        """Appends a random string of lowercase letters.

        Args:
            n (int): The number of characters to be appended.

        Returns: An AString instance

        """
        return self.append_random(string.ascii_lowercase, n)

    def remove_non_alphanumeric(self):
        """Removes non alpha-numeric characters from a string, including space and special characters.

        Returns: An AString with only alpha-numeric characters.

        """
        new_str = "".join([
            c for c in self
            if (c in string.digits or c in string.ascii_letters)
        ])
        return AString(new_str)

    def remove_non_ascii(self):
        """Removes non ASCII characters in the string.
        
        Returns: An AString with only ASCII characters.
        """
        return AString("".join([c for c in self if ord(c) <= 127]))

    def remove_escape_sequence(self):
        """Removes ANSI escape sequences, including color codes.

        Returns: An AString with escape sequence removed.

        """
        return AString(re.sub(r"\x1B\[[0-?]*[ -/]*[@-~]", "", self))


class FileName(AString):
    """Represents a filename and provides methods for modifying the filename.

    A "filename" is a string consist of a "basename" and an "extension".
    For example, in filename "hello_world.txt", "hello_world" is the basename and ".txt" is the extension.
    The extension can also be empty string. The filename should not contain a "." if the extension is empty.
    For example, "hello_world" is a filename with no extension.

    This class provides methods for modifying the "basename" of the filename.
    No modification will be applied to the "extension".

    Attributes:
        basename: The filename without extension
        extension: The file extension starting with "."

    Remarks:
        All methods will be operate on the "basename" only.
        This class is a sub-class of AString
        Most methods in this class support "Method Chaining", i.e. they return the FileName instance itself.
        If the method is inherited from AString or str, and the return value is also an AString or str
            the extension will be appended to the return value.
        If the return value of the method is not an AString or str,
            the extension will NOT be included in the return value.
        When using with operators (including +, *, slice, in),
            an FileName instance will be treated the same as str.
        len() will return the length of the whole filename, including "." and extension.

    """
    def __new__(cls, string_literal):
        name_splits = string_literal.rsplit('.', 1)
        filename = super(FileName, cls).__new__(cls, string_literal)
        filename.basename = name_splits[0]
        if len(name_splits) == 1:
            filename.extension = ""
        else:
            filename.extension = "." + name_splits[1]
        return filename

    def __getattribute__(self, item):
        """Wraps the existing methods of python AString to return FileName objects.
        """
        if item not in FileName.__dict__ and item in dir(AString):
            def method(s, *args, **kwargs):
                value = getattr(AString(self.basename), item)(*args, **kwargs)
                if isinstance(value, AString) or isinstance(value, str):
                    filename = type(s)(str(value) + self.extension)
                    return filename
                else:
                    return value
            # Bound method
            return method.__get__(self, type(self))
        else:
            # Delegate to parent
            return super(FileName, self).__getattribute__(item)

    @property
    def name_without_extension(self):
        """The filename without extension.
        """
        return os.path.basename(self.basename)

    def to_string(self):
        """Convert the FileName object to a string including basename and extension.
        """
        return str(self)


class Base64String(str):
    """Represents a string that is base64encoded from bytes.

    The error parameter used in the methods are passed into the str.encode() method.
    See Also: https://docs.python.org/3/library/stdtypes.html#str.encode
    
    """
    @staticmethod
    def encode_from_file(file_path, encoding='utf-8', errors="strict"):
        """Encodes a file into Base64 string.
        
        Args:
            file_path (str): Location of the file
            encoding (str, optional): Character encoding. Defaults to 'utf-8'.
            errors (str, optional): Error handling scheme. Defaults to "strict".
        
        Returns:
            Base64String: The encoded Base64 string.
        """
        with open(file_path, "rb") as f:
            content = f.read()
            return Base64String(base64.b64encode(content).decode(encoding, errors))

    @staticmethod
    def encode_from_string(s, encoding='utf-8', errors="strict"):
        """Encodes a string into Base64 string.
        
        Args:
            s (str): The string to be encoded.
            encoding (str, optional): Character encoding. Defaults to 'utf-8'.
            errors (str, optional): Error handling scheme. Defaults to "strict".
        
        Returns:
            Base64String: The encoded Base64 string.
        """
        return Base64String(base64.b64encode(s.encode(encoding, errors)).decode(encoding, errors))

    def decode_to_file(self, file_path, encoding='utf-8', errors="strict"):
        """Decodes the Base64 string and saves the contents into a file.
        
        Args:
            file_path (str): The location to save the file. Existing file will be overwritten.
            encoding (str, optional): Character encoding. Defaults to 'utf-8'.
            errors (str, optional): Error handling scheme. Defaults to "strict".

        """
        folder = os.path.dirname(file_path)
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(file_path, "wb") as f:
            f.write(base64.b64decode(self.encode(encoding, errors)))

    def decode_to_string(self, encoding='utf-8', errors="strict"):
        """Decodes the Base64 string and saves the contents into a string.
        
        Args:
            encoding (str, optional): Character encoding. Defaults to 'utf-8'.
            errors (str, optional): Error handling scheme. Defaults to "strict".
        
        Returns:
            [type]: [description]
        """
        return base64.b64decode(self.encode(encoding, errors)).decode(encoding, errors)


def stringify(obj):
    """Convert object to string.
    If the object is a dictionary-like object or list,
    the objects in the dictionary or list will be converted to strings, recursively.

    Returns: If the input is dictionary or list, the return value will also be a list or dictionary.

    """
    if isinstance(obj, abc.Mapping):
        obj = copy.deepcopy(obj)
        obj_dict = {}
        for key, value in obj.items():
            obj_dict[key] = stringify(value)
        return obj_dict
    elif isinstance(obj, list):
        str_list = []
        for item in obj:
            str_list.append(stringify(item))
        return str_list
    else:
        try:
            json.dumps(obj)
            return obj
        except TypeError:
            return str(obj)



import requests
import logging
import contextlib
from io import BytesIO
from .storage import StorageObject
from lxml import etree
from urllib import request
logger = logging.getLogger(__name__)


class WebAPI:
    """Provides method to access web API.

    This class uses python requests package.
    See https://2.python-requests.org/en/master/user/advanced/#request-and-response-objects

    Attributes:
        base_url: The base URL for all API endpoint.
        If base_url is specified, relative URL can be used to make requests.
        Relative URL will be appended to base URL when making the requests.
    
    """
    def __init__(self, base_url="", **kwargs):
        """Initializes API.

        Args:
            base_url: Base URL, the common URL prefix for all the API endpoints.
            **kwargs: keyword arguments to be encoded as GET parameters in the URL.
        """
        self.kwargs = kwargs
        self.headers = {}

        base_url = base_url
        if base_url.startswith("http://") or base_url.startswith("https://"):
            self.base_url = base_url
        else:
            raise ValueError("Base URL should start with http:// or https://")

    def add_header(self, **kwargs):
        """Adds a header to be used in all future HTTP requests

        Args:
            **kwargs: The key-value pairs to be added as the headers.

        """
        self.headers.update(kwargs)

    def request(self, method, url, **kwargs):
        """Sends HTTP request

        Args:
            method:
            url (str): The URL/Endpoint of the API.
                This can be a relative URL if base_url is specified in initialization.
            **kwargs: keyword arguments to be encoded as GET parameters in the URL.

        Returns:

        """
        url = self.build_url(url, **kwargs)
        method = str(method).lower()
        if not hasattr(requests, method):
            raise ValueError("Invalid method: %s" % method)
        request_func = getattr(requests, method)
        response = request_func(url, headers=self.headers)
        return response

    def get(self, url, **kwargs):
        """Makes a get request.
        Use keyword arguments to specify the query strings in the request.
        
        Args:
            url (str): The URL/Endpoint of the API.
                This can be a relative URL if base_url is specified in initialization.
            **kwargs: keyword arguments to be encoded as GET parameters in the URL.
        
        Returns: A Response Object
        """
        url = self.build_url(url, **kwargs)
        logger.debug("Requesting data from %s" % url)
        response = requests.get(url, headers=self.headers)
        logger.debug("Response code: %s" % response.status_code)
        if response.status_code != 200:
            logger.debug(response.content)
        return response

    def get_json(self, url, **kwargs):
        return self.get(url, **kwargs).json()

    def post(self, url, data, **kwargs):
        url = self.build_url(url, **kwargs)
        logger.debug("Posting data to %s" % url)
        response = requests.post(url, json=data, headers=self.headers)
        logger.debug("Response code: %s" % response.status_code)
        if response.status_code != 200:
            logger.debug(response.content)
        return response

    def post_json(self, url, data, **kwargs):
        return self.post(url, data, **kwargs).json()

    def delete(self, url, **kwargs):
        url = self.build_url(url, **kwargs)
        logger.debug("Deleting data from %s" % url)
        response = requests.delete(url, headers=self.headers)
        return response

    def build_url(self, url, **kwargs):
        """Builds the URL/Endpoint for a request.
        Keyword arguments are converted to query string in the URL.
        
        Args:
            url (str): The URL/Endpoint of the API.
            If url is relative url, it will be appended to the base_url.
            If url is absolute URL (starts with https:// or http://), the base_url will be ignored.
        
        Returns:
            str: The absolute URL/Endpoint of the API with query string.
        """
        url = url
        if not (url.startswith("http://") or url.startswith("https://")):
            url = "%s%s" % (self.base_url, url)
        query_dict = self.kwargs.copy()
        query_dict.update(kwargs)
        return self.append_query_string(url, **query_dict)
    
    @staticmethod
    def append_query_string(url, **kwargs):
        """Appends query string to a URL

        Query string is specified as keyword arguments.
        
        Args:
            url (str): URL
        
        Returns:
            str: URL with query string.
        """
        if kwargs and "?" not in url:
            url += "?"
        for key, val in kwargs.items():
            if "?" not in url:
                url += "?"
            if isinstance(val, list):
                pass
                url += "".join(["&%s=%s" % (key, v) for v in val])
            else:
                url += "&%s=%s" % (key, val)
        return url


class HTML:
    def __init__(self, uri):
        self.uri = uri
        self.__etree = None
        self.__content = None

    def read(self):
        obj = StorageObject(self.uri)
        if obj.scheme in ["http", "https"]:
            r = requests.get(self.uri)
            return r.content
        with open(self.uri, 'r') as f:
            return f.read()

    @property
    def content(self):
        if not self.__content:
            self.__content = self.read()
        return self.__content

    @property
    def etree(self):
        if not self.__etree:
            self.__etree = etree.parse(BytesIO(self.content), etree.HTMLParser())
        return self.__etree

    @staticmethod
    def __tags_to_list(parent, tag):
        elements = parent.findall(".//%s" % tag)
        if not elements:
            return None
        results = []
        for element in elements:
            text = element.text if element.text else ""
            results.append(text + ''.join(etree.tostring(e).decode() for e in element))
        return results

    @staticmethod
    def __append_data(to_list, parent, tag):
        data = HTML.__tags_to_list(parent, tag)
        if data:
            to_list.append(data)

    def get_tables(self):
        """Gets the data of of HTML tables in the web page as a list of dictionary.
        
        Returns:
            list: A list of dictionary, each contain data from a table in the web page.
            Each dictionary has two keys: "headers" and "data".
            Both "headers" and "data" are 2D lists.
        """
        html = self.etree
        html_tables = html.findall('.//table')
        data_tables = []
        
        for html_table in html_tables:
            table = {
                "headers": [],
                "data": []
            }
            rows = html_table.findall(".//tr")
            for row in rows:
                self.__append_data(table["headers"], row, "th")
                self.__append_data(table["data"], row, "td")
            data_tables.append(table)
        return data_tables


def download(url, file_path):
    """Downloads a file from a URL response.

    Args:
        url (str): The URL of the file to be downloaded.
        file_path (str): The path to store the file.

    Returns: None

    """
    url_response = request.urlopen(url)
    with open(file_path, 'wb') as out_file:
        with contextlib.closing(url_response) as fp:
            logger.debug("Downloading data from %s" % url)
            block_size = 1 << 16
            while True:
                block = fp.read(block_size)
                if not block:
                    break
                out_file.write(block)



import contextlib
import requests
import logging
from urllib import request, error
from .base import StorageIOBase
from io import UnsupportedOperation
logger = logging.getLogger(__name__)


class WebFile(StorageIOBase):
    def __init__(self, uri):
        self.response = None
        StorageIOBase.__init__(self, uri)

    def open(self, mode='r', closefd=True, opener=None):
        super().open(mode)
        # Follows redirect if there is any.
        if self.scheme.startswith("http"):
            response = requests.head(self.uri, allow_redirects=True)
            url = response.url
        else:
            url = self.uri
        logger.debug("Reading data from %s" % url)
        self.response = request.urlopen(url)
        return self

    def close(self):
        self._closed = True
        if self.response:
            self.response.close()
        self.response = None

    def read(self, size=None):
        return self.response.read(size)

    def write(self, b):
        raise UnsupportedOperation("Writing to HTTP file is not supported.")

    def delete(self):
        raise UnsupportedOperation("Deleting HTTP file is not supported.")

    def exists(self):
        try:
            response = request.urlopen(self.uri)
            if response.getcode() == 200:
                response.close()
                return True
        except error.URLError:
            pass
        return False

    @property
    def updated_time(self):
        return None



"""Contains tests for the file module.
"""
import logging
import os
import sys
try:
    from ..test import AriesTest
    from ..files import File, Markdown
except:
    aries_parent = os.path.join(os.path.dirname(__file__), "..", "..")
    if aries_parent not in sys.path:
        sys.path.append(aries_parent)
    from Aries.test import AriesTest
    from Aries.files import File, Markdown

logger = logging.getLogger(__name__)


class TestFile(AriesTest):
    def test_load_json(self):
        """Tests loading a json file
        """
        # File exists
        json_file = os.path.join(os.path.dirname(__file__), "fixtures", "test.json")
        json_dict = File.load_json(json_file)
        self.assertIn("key", json_dict)
        self.assertIn("list", json_dict)
        self.assertIn("dict", json_dict)
        # File does not exist, no default
        json_dict = File.load_json(os.path.join(os.path.dirname(__file__), "no_exist.json"))
        self.assertEqual(json_dict, {})
        # File does not exist, with default
        json_dict = File.load_json(
            os.path.join(os.path.dirname(__file__), "no_exist.json"),
            {"default_key": "default_value"}
        )
        self.assertEqual(json_dict["default_key"], "default_value")

    def test_file_signature(self):
        file_path = os.path.join(os.path.dirname(__file__), "fixtures", "excel_test_file.xlsx")
        f = File(file_path)
        self.assertIn("application/vnd.openxmlformats-officedocument", f.file_type())


class TestMarkdown(AriesTest):

    test_file = os.path.join(os.path.dirname(__file__), "fixtures", "test.md")

    def test_markdown_title(self):
        """Tests getting the title of a markdown file.
        """
        md = Markdown(self.test_file)
        self.assertEqual(md.title, "This is title")

    def test_markdown_links(self):
        expected_links = [
            '[an example](http://example.com/ "Title")',
            '[This link](http://example.net/)',
            '<a href="http://example.com/" title="Title">\nan example</a>',
            '<a href="http://example.net/">This link</a>',
            '[About](/about/)',
            '[id]: http://example.com/  "Optional Title Here"',
            '[foo]: http://example.com/  "Optional Title Here"',
            "[foo]: http://example.com/  'Optional Title Here'",
            '[foo]: http://example.com/  (Optional Title Here)',
            '[id]: <http://example.com/>  "Optional Title Here"',
            '[id]: http://example.com/longish/path/to/resource/here\n    "Optional Title Here"',
            '[Google]: http://google.com/',
            '[Daring Fireball]: http://daringfireball.net/',
            '[1]: http://google.com/        "Google"',
            '[2]: http://search.yahoo.com/  "Yahoo Search"',
            '[3]: http://search.msn.com/    "MSN Search"',
            '[google]: http://google.com/        "Google"',
            '[yahoo]:  http://search.yahoo.com/  "Yahoo Search"',
            '[msn]:    http://search.msn.com/    "MSN Search"',
            '<a href="http://google.com/"\ntitle="Google">Google</a>',
            '<a href="http://search.yahoo.com/" title="Yahoo Search">Yahoo</a>',
            '<a href="http://search.msn.com/" title="MSN Search">MSN</a>',
            '[Google](http://google.com/ "Google")',
            '[Yahoo](http://search.yahoo.com/ "Yahoo Search")',
            '[MSN](http://search.msn.com/ "MSN Search")'

        ]
        md = Markdown(os.path.join(os.path.dirname(__file__), "fixtures", "links.md"))
        matches = md.find_links()
        for i in range(len(expected_links)):
            self.assertEqual(expected_links[i], matches[i][0])
            # logger.debug("%s: %s" % (matches[i][0], matches[i][1]))



class DictList(list):
    """Represent a list of dictionaries
    """
    def unique_keys(self):
        """Gets the unique keys (union) of the dictionaries
        
        Returns:
            list: A list of keys from the dictionaries
        """
        keys = set()
        for d in self:
            for key in d.keys():
                keys.add(key)
        return keys

    def sort_by_value(self, key, reverse=False):
        """Sort the dictionaries by the value of a particular keys.
        
        Args:
            key: A key in the dictionaries.
            reverse (bool, optional): Reverse sorting. Defaults to False.
        
        Returns:
            list: A list of sorted dictionaries.
        """
        return sorted(self, key=lambda i : i.get(key))


def sort_lists(order_list, label_list, reverse=False):
    """Sort a two lists at the same time.
    The order will be determined by the order_list.

    Example:
        order_list = [2, 1, 3]
        label_list = ['A', 'B', 'C']
        After sorting,
        order_list = [1, 2, 3]
        label_list = ['B', 'A', 'C']
    
    Args:
        order_list (list): The list to be sorted.
        label_list (list): This list will be rearrange based on the order of order_list.
        reverse (bool, optional): Reverse sorting. Defaults to False.
    
    Returns:
        [type]: [description]
    """
    # Do nothing if the order_list if None or empty.
    if not order_list:
        return order_list, label_list
    order_list, label_list = (
        list(t) for t in zip(*sorted(zip(order_list, label_list), reverse=reverse))
    )
    return order_list, label_list



"""Contains classes for running functions/commands asynchronously.
"""
import logging
import pstats
import subprocess
import threading
import time
import traceback
from cProfile import Profile
# try:
from .outputs import CaptureOutput
# except SystemError:
#     import sys
#     from os.path import dirname
#     aries_parent = dirname(dirname(__file__))
#     if aries_parent not in sys.path:
#         sys.path.append(aries_parent)
#     from Aries.outputs import CaptureOutput

logger = logging.getLogger(__name__)


class Task:
    """A base class for representing a task like running a function or a command.

    Attributes:
        thread: The thread running the task, if the the task is running asynchronous.
            The thread value is set by run_async().
        The following attributes are designed to capture the output of running the task.
        std_out (str): Captured standard outputs.
        std_err (str): Captured standard errors.
        log_out (str): Captured log messages.
        exc_out (str): Captured exception outputs.
        returns: Return value of the task.
        pid (int): The PID of the process running the task.

    This class should not be initialized directly.
    The subclass should implement the run() method.
    The run() method should handle the capturing of outputs.
    """
    def __init__(self):
        self.pid = None
        self.thread = None
        self.returns = None
        self.exception = None
        self.std_out = ""
        self.std_err = ""
        self.log_out = ""
        self.exc_out = ""

    @property
    def log_list(self):
        """Log messages as a list.
        """
        return self.log_out.strip("\n").split("\n")

    def print_outputs(self):
        """Prints the PID, return value, stdout, stderr and logs.
        """
        print("=" * 80)
        print("PID: %s" % self.pid)
        print("RETURNS: %s" % self.returns)
        print("STD OUT:")
        for line in self.std_out.split("\n"):
            print(line)
        print("STD ERR:")
        for line in self.std_err.split("\n"):
            print(line)
        print("LOGS:")
        for line in self.log_out.split("\n"):
            print(line)

    def run(self):
        """Runs the task and capture the outputs.
        This method should be implemented by a subclass.
        """
        raise NotImplementedError(
            "A run() method should be implemented to run the task and capture the outputs."
        )

    def run_async(self):
        """Runs the task asynchronous by calling the run() method in a daemon thread.

        Returns: The daemon thread running the task.
        """
        thread = threading.Thread(
            target=self.run,
        )
        thread.daemon = True
        thread.start()
        self.thread = thread
        return self.thread

    def join(self):
        """Blocks the calling thread until the daemon thread running the task terminates.
        """
        if self.thread and self.thread.isAlive():
            return self.thread.join()
        else:
            return None


class FunctionTask(Task):
    """Represents a task of running a function.

    The return value of the function to be executed should be serializable.
    The logging will be captured by identifying the thread ID of the thread running the function.

    Attributes:
        thread: The thread running the function, if the the function is running asynchronous.
            The thread value is set by run_async().
        The following attributes are designed to capture the output of running the function.
        std_out (str): Captured standard outputs.
        std_err (str): Captured standard errors.
        log_out (str): Captured log messages.
        exc_out (str): Captured exception outputs.
        returns: Return value of the task.
        pid (int): The PID of the process running the task.

        func: The function to be executed.
        args: The arguments for executing the function.
        kwargs: The keyword arguments for executing the function.

    Remarks:
        std_out and std_err will contain the outputs from all threads running in the same process.

    """
    # Stores a list of attribute names to be captured from the process running the function
    __output_attributes = [
        "std_out",
        "std_err",
        "log_out",
        "exc_out",
        "returns",
    ]

    def __init__(self, func, *args, **kwargs):
        """Initializes a task to run a function.

        Args:
            func: The function to be executed.
            *args: A list of arguments for the function to be executed.
            **kwargs: A dictionary of keyword arguments for the function to be executed.
        """
        super(FunctionTask, self).__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs

        self.log_filters = []

        self.out = None

    def add_log_filter(self, log_filter):
        self.log_filters.append(log_filter)
        return self

    def __unpack_outputs(self, out):
        """Sets a list of attributes (self.__output_attributes) by copying values from a dictionary.

        Args:
            out (dict): The dictionary containing the values for attributes.
                The keys in the dictionary must be the same as the attribute names.

        """
        for k in self.__output_attributes:
            setattr(self, k, out.get(k))

    def __pack_outputs(self, out):
        """Saves a list of attributes (self.__output_attributes) to a dictionary.

        Args:
            out: An object with all attributes listed in self.__output_attributes.

        """
        return {
            k: getattr(out, k) for k in self.__output_attributes
        }

    def __run(self):
        try:
            with CaptureOutput(filters=self.log_filters) as out:
                self.out = out
                # Returns may not be serializable.
                out.returns = self.func(*self.args, **self.kwargs)
        except Exception as ex:
            # Catch and save the exception
            # run() determines whether to raise the exception
            #   base on "surpress_exception" argument.
            self.exception = ex
        else:
            # Reset self.exception if the run is successful.
            # This is for run_and_retry()
            self.exception = None
        try:
            # name = self.func.__name__ if hasattr(self.func, "__name__") else str(self.func)
            # logger.debug("Finished running %s()..." % name)
            return self.__pack_outputs(out)
        except Exception as ex:
            print(ex)
            return {
                "exc_out": traceback.format_exc()
            }

    def exit_run(self):
        """Additional processing before exiting the task.

        This method is intended to be implemented by a subclass.
        """
        pass

    def run(self, suppress_exception=True):
        """Runs the function and captures the outputs.
        """
        # receiver, pipe = Pipe()
        # p = Process(target=self.__run, args=(pipe,))
        # p.start()
        #
        # self.pid = p.pid
        # print("%s PROCESS STARTED" % p.pid)
        # out = receiver.recv()
        # print("%s MESSAGE RECEIVED" % p.pid)
        # p.terminate()
        self.__unpack_outputs(self.__run())
        if self.exc_out:
            print(self.exc_out)
        self.exit_run()
        if not suppress_exception and self.exception is not None:
            raise self.exception
        return self.returns

    def run_profiler(self):
        """Runs the function with profiler.
        """
        profile = Profile()
        profile.runcall(self.func, *self.args, **self.kwargs)
        stats = pstats.Stats(profile)
        stats.strip_dirs()
        # Display profiling results
        stats.sort_stats('cumulative', 'time').print_stats(0.1)

    def run_and_retry(self, max_retry=10, exceptions=Exception, 
                      base_interval=2, retry_pattern='exponential'):
        """Runs the function and retry a few times if certain exceptions occurs.
        The time interval between the ith and (i+1)th retry is base_interval**i, 
            i.e. interval increases exponentially.

        Args:
            max_retry (int): The number of times to re-try.
            exceptions: An exception class or A tuple of exception classes.
            base_interval (int): The interval before the first retry in seconds.
            retry_pattern (str): The pattern of the retry interval. 
                "exponential": The time between two retries will increase exponentially.
                    i.e., the interval will be "base_interval ** i" after the ith try.
                "linear": The time between two retries will increase linear.
                    i.e., the interval will be "base_interval * i" after the ith try.

        Returns: The return value of the function.

        """
        error = None
        for i in range(max_retry):
            try:
                results = self.run(suppress_exception=False)
            except exceptions as ex:
                error = ex
                traceback.print_exc()
                if retry_pattern == "exponential":
                    time.sleep(base_interval ** (i + 1))
                else:
                    time.sleep(base_interval * (i + 1))
            else:
                return results
        # The following will be executed only if for loop finishes without break/return
        else:
            raise error


class ShellCommand(Task):
    """Represents a task of running a shell command.

    Attributes:
        thread: The thread running the task, if the the task is running asynchronous.
            The thread value is set by run_async().
        The following attributes are designed to capture the output of running the task.
        std_out (str): Captured standard outputs.
        std_err (str): Captured standard errors.
        log_out (str): Captured log messages.
        exc_out (str): Captured exception outputs.
        returns: Return value of the task.
        pid (int): The PID of the process running the task.
    
    This class can be used to run a shell command and capture the outputs.
    For example, the command "ls -a ~" displays all files in the user's home directory.
    The following code runs this command:

        cmd = "ls -a ~"
        task = ShellCommand(cmd)
        task.run()
        print(task.std_out)
    
    The outputs are stored in "task.std_out" as a string.

    Run the command asynchronously, if the command takes a long time to complete:

        cmd = "YOUR_AWESOME_COMMAND"
        task = ShellCommand(cmd)
        task.run_async()
        # Feel free to do something else here
        # ...
        # Get the outputs
        task.join()
        print(task.std_out)
    
    """
    def __init__(self, cmd):
        super(ShellCommand, self).__init__()
        self.cmd = cmd
        self.process = None

    def run(self):
        """Runs the command with Popen()
        """
        self.process = subprocess.Popen(
            self.cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True
        )
        self.pid = self.process.pid
        out, err = self.process.communicate()
        self.std_out = out.decode()
        self.std_err = err.decode()
        self.returns = self.process.returncode
        return self
